type RKOMMarket {
    name: String
    timezone: String
    startOfWeek: Int
}


type ReserveScenario{
    volume: Int
    auction: String
    product: String
    block: String
    reserveGroup: String
}


type RKOMBid {
    name: String
    date: [DateTransformation]
    market: RKOMMarket
    method: String
    minimumPrice: Float
    pricePremium: Float
    priceScenarios: [ScenarioMapping]
    reserveScenarios: [ReserveScenario]
    watercourse: String
}

type RKOMProcess {
    name: String
    bid: RKOMBid
    shop: ShopTransformation
    processEvents: [String]
    timezone: String
    plants: [String]
    incremental_mappings: [ScenarioMapping]
}

type RKOMCombinationBid{
    name: String
    auction: String
    rkomBidConfigs: [String]
}

type RKOMBidCombination {
    name: String
    auction: String
    bid: RKOMCombinationBid
}
