type InputTimeSeriesMapping {
  shopObjectType: String
  shopObjectName: String
  shopAttributeName: String
  cdfTimeSeries: TimeSeries
  retrieve: String
  transformations: [ValueTransformation]
  aggregation: String
}
