type ProductionPlanTimeSeries {
    name: String
    series: [TimeSeries]
}

type BenchmarkBid{
    name: String
    date: [DateTransformation]
    market: NordPoolMarket
}

type BenchmarkProcess{
    name: String
    productionPlanTimeSeries: [ProductionPlanTimeSeries]
    metrics: JSONObject
    bid: BenchmarkBid
    shop: ShopTransformation
    runEvents: [String]
}
